local Harvest = _G["Harvest"]
local CallbackManager = Harvest.callbackManager
local Events = Harvest.events

local PinsShared = ZO_Object:Subclass()
Harvest.PinsShared = PinsShared

function PinsShared:New(...)
	local instance = {}
	self.__index = self
	setmetatable(instance, self)
	instance:Initialize(...)
	return instance
end

function PinsShared:Initialize()
	self.pinTable = {}
	self.hasEntries = false
	self.lastViewedX = -10
	self.lastViewedY = -10
	CallbackManager:RegisterForEvent(Events.CHANGED_NODE_HIDDEN_STATE, function(event, map, nodeId, newState)
		self:OnChangedNodeHiddenState(map, nodeId, newState)
	end)
	local callback = function(...) self:OnNodeChangedCallback(...) end
	CallbackManager:RegisterForEvent(Events.NODE_DELETED, callback)
	CallbackManager:RegisterForEvent(Events.NODE_UPDATED, callback)
	CallbackManager:RegisterForEvent(Events.NODE_ADDED, callback)
end

function PinsShared:OnNodeChangedCallback(event, map, pinTypeId, nodeId)
	local nodeAdded = (event == Events.NODE_ADDED)
	local nodeUpdated = (event == Events.NODE_UPDATED)
	local nodeDeleted = (event == Events.NODE_DELETED)
	
	if not self:CanRefresh() then
		return
	end

	if not self:IsActiveMap(map) then
		return
	end

	if self:ShouldRefreshAllPins() then
		self:RefreshPins(pinTypeId)
	else
		-- no addon is used, so we can refresh a single pin by removing and recreating it
		if not nodeAdded then
			self:RemoveNode(nodeId)
		end
		-- the (re-)creation of the pin is performed
		if not nodeDeleted and not self.mapCache:IsHidden(nodeId) then
			self:AddNode(nodeId)
		end
	end
end

function PinsShared:OnChangedNodeHiddenState(map, nodeId, newState)
	if self:IsActiveMap(map) then
		if newState then
			self:RemoveNode(nodeId)
		else
			self:AddNode(nodeId)
		end
	end
end

function PinsShared:RemovePins(pinTypeId)
	if pinTypeId then
		self:Clear(pinTypeId)
		self.pinManager:RemovePins(Harvest.GetPinType(pinTypeId))
	else
		self:Clear()
		for _, pinTypeId in ipairs(Harvest.PINTYPES) do
			if pinTypeId ~= Harvest.TOUR then
				self.pinManager:RemovePins(Harvest.GetPinType(pinTypeId))
			end
		end
	end
end

function PinsShared:RefreshPins(pinTypeId)
	if pinTypeId then
		if pinTypeId == Harvest.TOUR then
			self.library:RefreshPins( Harvest.GetPinType(pinTypeId) )
			return
		end
		
		self.library:RefreshPins(Harvest.GetPinType(pinTypeId))
		self:RemovePins(pinTypeId)

		if not Harvest.IsUpdateQueueEmpty() then
			Harvest.Debug("your data is still being refactored/updated" )
			return
		end

		if not self:CanRefresh() then
			return
		end
		
		if Harvest.IsMapPinTypeVisible(pinTypeId) then
			self:AddVisibleNodesOfPinType(pinTypeId)
		end
	else
		for _, pinTypeId in pairs(Harvest.PINTYPES) do
			self.library:RefreshPins(Harvest.GetPinType(pinTypeId))
		end
		self.library:RefreshPins( Harvest.GetPinType( "0" ))
	end
end

function PinsShared:UpdateVisibleMapPins()
	if not self.CanRefresh() then return end

	local map = Harvest.GetMap()
	local x, y = GetMapPlayerPosition("player")

	self:AddAndRemoveVisblePins(map, x, y)
end

function PinsShared:PinTypeRefreshCallback()
	Harvest.Debug("Refresh of pins requested")

	-- stop adding/removing map pins as they belong to the previous map
	self:RemovePins()

	-- data is still being manipulated, better if we don't access it yet
	if not Harvest.IsUpdateQueueEmpty() then
		Harvest.Debug("your data is still being refactored/updated" )
		return
	end

	if not self:CanRefresh() then
		Harvest.Debug("pins could not be refreshed" )
		return
	end

	local map, x, y, measurement, zoneIndex = Harvest.GetLocation( true )
	self:SetToMapAndPosition(map, x, y, measurement, zoneIndex)
	for _, pinTypeId in ipairs(Harvest.PINTYPES) do
		if Harvest.IsMapPinTypeVisible(pinTypeId) then
			self:AddVisibleNodesOfPinType(pinTypeId)
		end
	end

end

function PinsShared:AddNode(nodeId)
	if not self.mapCache then return end
	if self.pinTable[nodeId] and self.pinTable[nodeId] < 0 then
		self.pinTable[nodeId] = nil
	else
		self.pinTable[nodeId] = self.mapCache.pinTypeId[nodeId]
	end
	self.hasEntries = (next(self.pinTable) ~= nil)
end

function PinsShared:RemoveNode(nodeId)
	if not self.mapCache then return end
	self.pinTable[nodeId] = nil
	self.hasEntries = (next(self.pinTable) ~= nil)

	self:RemovePin(self.mapCache.pinTypeId[nodeId], nodeId)
end

-- If the player moved, some pins enter the visible radius while others leave it.
-- This function updates the queue with new creation/removal commands according to the movement.
function PinsShared:AddAndRemoveVisblePins(map, x, y)
	if not self.mapCache then return end
	-- abort if the maps don't match
	-- i.e this queue is used by the compass and the player is viewing a different map than he is currently on.
	if self.mapCache.map ~= map then return end

	local shouldSaveCoords
	shouldSaveCoords = self.mapCache:SetPrevAndCurVisibleNodesToTable(self.lastViewedX, self.lastViewedY, x, y, self.pinTable)

	if shouldSaveCoords then
		self.lastViewedX = x
		self.lastViewedY = y
	end

	self.hasEntries = (next(self.pinTable) ~= nil)
end

function PinsShared:SetToMapAndPosition(map, x, y, measurement, zoneIndex)
	if self.mapCache then
		self.mapCache.accessed = self.mapCache.accessed - 1
	end
	self.currentMap = map
	self.mapCache = Harvest.Data:GetMapCache(nil, map, measurement, zoneIndex)
	if not self.mapCache then
		self.hasEntries = false
		return
	end
	self.mapCache.accessed = self.mapCache.accessed + 1
	self.lastViewedX = x
	self.lastViewedY = y
end

---Adds only nodes of the given pinType to the creation queue. Used when only a certain pinType is refreshed.
-- (E.g. the player enabled the pinType on the map)
function PinsShared:AddVisibleNodesOfPinType(pinTypeId)
	if not self.mapCache then return end
	Harvest.Data:CheckPinTypeInCache(pinTypeId, self.mapCache)
	
	self.mapCache:GetVisibleNodes(self.lastViewedX, self.lastViewedY, pinTypeId, self.pinTable)
	self.hasEntries = (next(self.pinTable) ~= nil)
end

-- Clear the queue. Used when the currently viewed map changes or the pins are to be refreshed.
function PinsShared:Clear(pinTypeId)
	if not pinTypeId then
		self.pinTable = {}
		self.hasEntries = false
	else
		for nodeId, pinType in pairs(self.pinTable) do
			if pinType == pinTypeId or pinType == -pinTypeId then
				self.pinTable[nodeId] = nil
			end
		end
		self.hasEntries = (next(self.pinTable) ~= nil)
	end
end

-- Creates and removes a bunch of queued pins.
function PinsShared:PerformActions()
	if not self.hasEntries then
		return
	end
	if self.ShouldDelay() then
		return
	end
	local numNodes = Harvest.GetDisplaySpeed()
	local x, y
	for nodeId, pinTypeId in pairs(self.pinTable) do
		if pinTypeId > 0 then
			x = self.mapCache.localX[nodeId]
			y = self.mapCache.localY[nodeId]
			self:CreatePin(pinTypeId, nodeId, x, y)
		else
			x = self.mapCache.localX[nodeId]
			y = self.mapCache.localY[nodeId]
			self:RemovePin(-pinTypeId, nodeId, x, y)
		end
		self.pinTable[nodeId] = nil
		numNodes = numNodes - 1
		if numNodes <= 0 then
			return
		end
	end
	self.hasEntries = false
end

function PinsShared:IsActiveMap(map)
	if not self.mapCache then return false end
	return (self.mapCache.map == map)
end
